import os
import re
import json
import random
import string
import shutil
from pathlib import Path
from typing import Literal, List
import GPUtil

import cudf
import pandas as pd
import numpy as np

import dask.dataframe as dd
import dask.config
dask.config.set({'dataframe.backend':'cudf'})
from dask.diagnostics import ProgressBar
import rmm

from ..compute.backend import infer_cuda
from ..utils import as_path

def split_into_groups(series, cutoff):
    groups = []
    
    while len(series.index) > 0:
        current_group = []
        current_sum = 0
        for username, storage_size in series.items():
            if storage_size > cutoff:
                groups.append({username})
                series = series.drop(username)
                break
            elif current_sum + storage_size <= cutoff:
                current_group.append(username)
                current_sum += storage_size
        
        series = series.drop(current_group)
        if current_group:
            groups.append(set(current_group))
    
    return groups

def calc_tld_mem(df):
    mem = df.groupby('tld',observed=True).apply(lambda x: x.memory_usage(deep=True).sum())
    return mem

def define_tld_groups(input,cutoff):
    ddf = dd.read_parquet(input,columns = ['size','kballoc','access','create','modify','uid','gid','path','tld'])
    with ProgressBar():
        tld_mem = ddf.map_partitions(calc_tld_mem).compute()
    tld_mem = tld_mem.groupby(tld_mem.index).sum().divide(1024**3).to_pandas()
    grps = split_into_groups(tld_mem,cutoff)
    return grps

def nested_list_to_log(nest,file):
    """
    Writes a list of lists to a text log

    Args:
    nest (list): A list of lists to be converted.
    """
    with open(file, 'w', newline='') as f:
        for l in nest:
            f.write(f"{','.join(l)}\n")

def prep_hivize(args):
    grps = define_tld_groups(args.get('parquet_path'),args.get('cutoff'))    
    misc_path = args.get('parquet_path').parent.joinpath('misc','tld_grps.txt')
    misc_path.parent.mkdir(exist_ok = True, parents = True)
    nested_list_to_log(grps,misc_path)
    ngroups = len(grps)
    grp_file = str(misc_path)
    args.update({'ngroups':ngroups,
                'grp_file':grp_file})
    return args

def hivize(
        parquet_path: str | Path, 
        hive_path: str | Path, 
        tld: str | List[str] | None = None,
        staging_path: str | Path | None = None,
        partition_size: str = '100MiB', 
        with_cuda: bool | Literal['infer'] = 'infer',
        write_metadata: bool = True,
        **kwargs
    ) -> None:
    parquet_path = as_path(parquet_path).resolve()
    hive_path = as_path(hive_path).resolve()
    
    if staging_path is None:
        rand_str = ''.join(random.choices(string.ascii_letters + string.digits, k=8))
        staging_path = Path(os.getenv('TMPDIR')).joinpath(os.getenv('USER'),f'hive-{rand_str}')
        print(f"INFO: Using {staging_path} as temporary directory",flush=True)
    else:
        staging_path = as_path(staging_path)

    hive_path.mkdir(exist_ok=True,parents=True)
    staging_path.mkdir(exist_ok=True,parents=True)

    if with_cuda == 'infer':
        with_cuda = infer_cuda()

    if with_cuda:
        import dask_cudf as backend
        from dask_cudf.core import from_cudf as from_local
        dask.config.set({'dataframe.backend':'cudf'})
        rmm.reinitialize(
            pool_allocator=True,
            managed_memory=True,
            initial_pool_size='60GiB'
        )
    else:
        import dask as backend
        from dask.dataframe import from_pandas as from_local
        dask.config.set({'dataframe.backend':'pandas'})

    def indexed_name(ind):
        return f"indexed-{ind}.parquet"

    if tld is not None:
        if not isinstance(tld,list):
            tld = [tld]
        predicates = [('tld','in',tld)]
    else:
        predicates = None
    print(f"DEBUG: Filtering predicates are: {predicates}",flush=True)

    acq = re.search(r'[\d]{4}-[\d]{2}-[\d]{2}',str(parquet_path)).group(0)
    print(f"DEBUG: Acquisition date is {acq}",flush=True)

    # The flat parquet is initially read in via dask to avoid reading the full dataset into memory which happens even 
    # when including predicates for filtering. The dask dataframe is converted to a regular dataframe to drastically 
    # improve indexing and sorting by removing partitions. The sorted dataframe is converted back to a dask dataframe 
    # to create partitions within the parquet dataset and write to multiple files defined by those partitions.
    ddf = backend.read_parquet(parquet_path,filters=predicates,columns = ['size','kballoc','access','create','modify','uid','gid','path','tld'])
    df = ddf.compute()
    df = df.set_index('path').sort_index().assign(acq=acq)
    ddf = from_local(df).repartition(partition_size=partition_size,force=True)
    ddf.to_parquet(staging_path,partition_on=['tld','acq'],name_function = indexed_name)

    if write_metadata:
        (
            df
            .groupby(['tld','acq'])
            .apply(lambda x: 
                   write_dataset_metadata(
                       x, 
                       staging_path.joinpath(f"tld={x['tld'].iloc[0]}",f"acq={x['acq'].iloc[0]}")
                    )
                )
        )

    shutil.copytree(staging_path,hive_path,dirs_exist_ok=True)
    shutil.rmtree(staging_path)
    pass

def bytes_to_human_readable_size(num_bytes):
    units = ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB']
    
    # Handle the case where num_bytes is 0
    if num_bytes == 0:
        return "0 B"
    
    # Calculate the appropriate unit. Take the floor if the number is not divisible by 1024
    unit_index = min(len(units) - 1, int(np.log2(num_bytes) // 10))
    num = int(num_bytes // (1024 ** unit_index))
    
    # Format the number to 2 decimal places and append the unit
    return f"{num} {units[unit_index]}"

def create_bin_labels(bins, include_outer_bins = True):
    if any([b < 0 for b in bins]):
        raise ValueError("File size cannot be less than 0 B")
    bins = list(set(bins))
    bins.sort() # Sorts and removes any duplicates
    if np.isinf(bins[-1]):
        bins.remove(np.inf)
    
    labels = []
    
    # Check for including a bin for all files less than the smallest value if 
    if include_outer_bins and bins[0] != 0:
        labels.append(f"0 B-{bytes_to_human_readable_size(bins[0])}")
    
    for i in range(len(bins) - 1):
        lower_bound = bytes_to_human_readable_size(bins[i])
        upper_bound = bytes_to_human_readable_size(bins[i + 1])
        labels.append(f"{lower_bound}-{upper_bound}")
    
    if include_outer_bins:
        labels.append(f">{bytes_to_human_readable_size(bins[-1])}")
    return labels,bins

def calculate_size_distribution(
        sizes: pd.Series | cudf.Series, 
        size_bins: List[int | float] = [0,4096] + [x*1024**3 for x in [1,10,50,100]] + [1024**4],
        include_outer_bins: bool = True,
        **kwargs
) -> pd.Series | cudf.Series:
    
    size_labels,size_bins = create_bin_labels(size_bins,include_outer_bins)

    if include_outer_bins:
        if size_bins[0] < 0:
            size_bins.insert(0,0)
        if float("inf") not in size_bins:
            size_bins.append(float("inf"))

    fn = pd.cut if isinstance(sizes,pd.Series) else cudf.cut
    size_grps = fn(sizes,bins=size_bins,labels=size_labels,ordered=True,include_lowest=True,**kwargs)

    return size_grps

def write_dataset_metadata(df: cudf.DataFrame | pd.DataFrame, parquet_path: str | Path) -> dict:
    parquet_path = as_path(parquet_path)
    if isinstance(df,cudf.DataFrame):
        vram_usage = df.memory_usage(deep=True).to_dict()
        ram_usage  = df.to_pandas().memory_usage(deep=True).to_dict()
    else:
        if len(GPUtil.getAvailable()) > 0:
            vram_usage = cudf.from_pandas(df).memory_usage(deep=True).to_dict()
        else:
            vram_usage = 'NA'
        ram_usage = df.memory_usage(deep=True).to_dict()
    
    df = df[['size']]
    df['grp'] = calculate_size_distribution(df['size'])
    
    size_dist = (
        df
        .groupby('grp',observed=True)['size']
        .agg(['count','sum'])
        .rename(columns={'count':'file_count','sum':'bytes'})
        .T
        .to_dict()
    )

    metadata = {
        'num_rows': df.shape[0],
        'num_columns': df.shape[1],
        'vram_usage': vram_usage,
        'ram_usage': ram_usage,
        'size_distribution': size_dist
    }

    with open(parquet_path.joinpath('_metadata.json'),'w') as f:
        json.dump(metadata,f)